/* mkdocstrings styling */

.doc.doc-heading {
    padding-left: 1ch;
    padding-bottom: 5px;
    background-color: var(--md-code-bg-color);
}

div.doc-contents:not(.first) {
    padding-left: 2ch;
}

/* Class name headings */
h3.doc.doc-heading>code {
    font-weight: bold;
}

/* Class attribute and method headings */
h5.doc.doc-heading {
    font-size: 1em;
    text-transform: none;
}

/* Bold function, method, and attribute names */
.doc.doc-heading span {
    font-weight: bold;
}

.doc.doc-heading span~span {
    font-weight: normal;
}
